import torch 
from torch.distributions.normal import Normal 
from torch.distributions.multivariate_normal import MultivariateNormal
from torch.distributions.uniform import Uniform

import math
import numpy
import matplotlib.pyplot as plt
import numpy as np

def Simulator(simulatorName):

    if(simulatorName == 'Gaussian'):
        return GaussianSimulator()
    elif(simulatorName == 'simple2dGaussian'):
        return simple2dGaussianSimulator()
    elif(simulatorName == 'cubic2dGaussianSimulator'):
        return cubic2dGaussianSimulator()
    elif(simulatorName == '2dMoon'):
        return MoonSimulator()
    elif(simulatorName == 'IK'):
    	return InverseKinematicsSimulator()
    else: # TODO: throw error
        return None

class GaussianSimulator():
    """
    Simulator defined in: Sequential Neural Likelihood; Fast Likelihood-free Inference 
                          with Autoregressive Flows
                          George Papamakarios, David C. Sterratt, Iain Murray
                          21 Jan 2019
    """

    def __init__(self):
        self.ydim = 8
        self.xdim = 5

    def sample_prior(self, size):
        uniform = Uniform(torch.zeros(size, 5)+ torch.tensor([-3.]), torch.zeros(size,5)+torch.tensor([3.]))
        return uniform.sample()

    def log_prior(self, x):
        assert x.shape[0] == 5, 'At the moment, only one x is supported'
        uniform = Uniform(torch.tensor([-3.]), torch.tensor([3.]))
        return uniform.log_prob(x).sum(dim=0)
    
    def corrupt(self, x):
        means = x[:,:2]
        s1 = torch.pow(x[:,2],2)
        s2 = torch.pow(x[:,3],2)
        pho = torch.tanh(x[:,4])
        
        cov = torch.zeros(x.shape[0], 2, 2)
        cov[:, 0, 0] = torch.pow(s1,2)
        cov[:, 0, 1] = pho * s1 * s2
        cov[:, 1, 0] = pho * s1 * s2
        cov[:, 1, 1] = torch.pow(s2,2)
        
        normal = MultivariateNormal(means, cov)
        
        Y = torch.zeros(x.shape[0],8)
        Y[:,:2] = normal.sample()
        Y[:, 2:4] = normal.sample()
        Y[:, 4:6] = normal.sample()
        Y[:, 6:] = normal.sample()
        
        return Y

    def log_likelihood(self, x, y):

    	assert x.shape[0]==5, 'At the moment, only one x is supported'

    	means = x[:2]
    	s1 = torch.pow(x[2],2)
    	s2 = torch.pow(x[3],2)
    	pho = torch.tanh(x[4])

    	cov = torch.zeros(1, 2, 2)
    	cov[:, 0, 0] = torch.pow(s1,2)
    	cov[:, 0, 1] = pho * s1 * s2
    	cov[:, 1, 0] = pho * s1 * s2
    	cov[:, 1, 1] = torch.pow(s2,2)

    	normal = MultivariateNormal(means, cov)
    	return normal.log_prob(y[:2]) + normal.log_prob(y[2:4]) + normal.log_prob(y[4:6]) + normal.log_prob(y[6:])

    def get_MLE(self, observation):
        """
        Given an observation, returns the most likely parameter.
        """
        
        y1 = observation[0, 0::2]
        y2 = observation[0, 1::2]
        
        x1_mle = y1.mean().item()
        x2_mle = y2.mean().item()
        
        obs = torch.cat((y1.unsqueeze(0), y2.unsqueeze(0)), dim=0)
        covariance = numpy.cov(obs.data.cpu().numpy())
        
        s1 = numpy.sqrt(covariance[0, 0])
        s2 = numpy.sqrt(covariance[1, 1])
        pho = covariance[0, 1] /(s1 * s2)
        
        x3_mle = numpy.sqrt(s1)
        x4_mle = numpy.sqrt(s2)
        x5_mle = numpy.arctanh(pho)
        
        return torch.tensor([x1_mle, x2_mle, x3_mle, x4_mle, x5_mle])

class simple2dGaussianSimulator():
    def __init__(self, means_prior=[3.0, 7.0]):
        self.means_prior = means_prior
        self.ydim = 2
        self.xdim = 2

    def sample_prior(self, size):
    	return sample_multivariate_gaussian(size, self.means_prior[0], self.means_prior[1])

    def log_prior(self, x):
        assert x.shape[0] == 2, 'At the moment, only one x is supported'
        m = MultivariateNormal(torch.tensor(self.means_prior), torch.eye(2))
        return m.log_prob(x)

    def log_likelihood(self, x, y):
        assert y.shape[0] == x.shape[0] == 2, 'At the moment, only one data point is supported'
        m = MultivariateNormal(2 * x + 1, torch.eye(2))
        return m.log_prob(y)


    def corrupt(self, x):
        means =  2 * x  + 1
        cov = torch.eye(2)
        m = MultivariateNormal(means, cov)
        return m.sample()

    def get_MLE(self, observation):
        """
        Given an observation, returns the most likely parameter.
        """
        return (observation - 1)/2


class cubic2dGaussianSimulator():
    def __init__(self, means_prior=[3.0, 7.0]):
        self.means_prior = means_prior

    def sample_prior(self, size):
    	return sample_multivariate_gaussian(size, self.means_prior[0], self.means_prior[1])

    def corrupt(self, x):
        means = (0.3 * x**3 + 0.2 * x**2 + x  + 1)
        cov = torch.eye(2)
        m = MultivariateNormal(means, cov)
        return m.sample()

    def get_MLE(self, observation):
        """
        Given an observation, returns the most likely parameter.
        """
        raise NotImplementedError("Not implemented yet")


class MoonSimulator():
    '''
    Simulator defined in: Automatic Posterior Transformation for Likelihood-free Inference
                		  David S. Greenberg, Marcel Nonnenmacher, Jakob H. Macke
                		  17 May 2019
    '''
    def __init__(self, mean_radius=0.1, std_radius = 0.01):
        self.mean_radius = mean_radius
        self.std_radius = std_radius
        self.ydim = 2
        self.xdim = 2

        

    def sample_prior(self, size):
        m = Uniform(torch.zeros((size,2)) -1., torch.zeros((size,2)) + 1.) 
        theta = m.sample()
        x = theta
        return x

    def log_prior(self, x):
        assert x.shape[0] == 2, 'At the moment, only one x is supported'
        uniform = Uniform(torch.tensor([-1.]), torch.tensor([1.]))
        return uniform.log_prob(x).sum(dim=0)

    def corrupt(self, x):

        theta = x
        size_theta = theta.shape
    
        uniform = Uniform(torch.zeros(size_theta[0])+ torch.tensor([-math.pi/2]), torch.zeros(size_theta[0])+torch.tensor([math.pi/2]))
        normal = Normal(torch.zeros(size_theta[0])+ self.mean_radius,torch.zeros(size_theta[0]) + self.std_radius)

        a = uniform.sample()
        r = normal.sample()
        
        p = torch.empty(size_theta)

        p[:,0] = torch.mul(r, torch.cos(a)) + 0.25
        p[:,1] = torch.mul(r, torch.sin(a))
        
        q = torch.empty(size_theta)
        q[:,0] = -torch.abs(theta[:,0] + theta[:,1])/math.sqrt(2)
        q[:,1] = (-theta[:,0] + theta[:,1])/math.sqrt(2)
        
        x = p + q

        y = x
        return y

    def likelihood(self, y):
        '''
        returns the likelihood of the corrupted data y
        '''
        raise NotImplementedError("No analytic formula known")

    def get_MLE(self, observation):
        """
        Given an observation, returns the most likely parameter.
        """
        raise NotImplementedError("Not implemented yet")

class InverseKinematicsSimulator():
    """
    Simulator used in: "Analyzing Inverse Problems with Invertible Neural Networks"
                        Ardizzone1 et al.
                        Feb 2019
    """
    
    def __init__(self, prior_var=[1/16, 1/4, 1/4, 1/4], l1=1/2, l2=1/2, l3=1, noise=0.01/180*np.pi):
        self.prior_var = torch.tensor(prior_var)
        self.l1=l1
        self.l2=l2
        self.l3=l3
        self.noise = noise
        self.ydim = 2
        self.xdim = 4
    
    def sample_prior(self, size):
        prior = MultivariateNormal(torch.zeros(size, 4), self.prior_var * torch.eye(4))
        return prior.sample()

    def log_prior(self, x):
        assert x.shape[0] == 4, 'At the moment, only one x is supported'
        prior = MultivariateNormal(torch.zeros(4), self.prior_var * torch.eye(4))
        return prior.log_prob(x)
    
    def corrupt(self, x):
        
        y = torch.empty((x.shape[0], 2))

        noise_distribution = Normal(torch.zeros(x.shape[0]), self.noise)
        
        y[:,0] = self.l1 * torch.sin(x[:,1] + noise_distribution.sample()) + self.l2 * torch.sin(x[:,2] + x[:,1] + noise_distribution.sample()) + self.l3 * torch.sin(x[:,3] + x[:,1] + x[:,2] + noise_distribution.sample()) + x[:,0]
        y[:,1] = self.l1 * torch.cos(x[:,1] + noise_distribution.sample()) + self.l2 * torch.cos(x[:,2] + x[:,1] + noise_distribution.sample()) + self.l3 * torch.cos(x[:,3] + x[:,1] + x[:,2] + noise_distribution.sample())
        
        return y

    def likelihood(self, y):
        '''
        returns the likelihood of the corrupted data y
        '''
        raise NotImplementedError("No analytic formula known to my knownledge")

    def get_MLE(self, observation):
        """
        Given an observation, returns the most likely parameter.
        """
        raise NotImplementedError("Not implemented yet")
        
    def make_plot(self, x, y=None, savepath=None, do_show=True, disable_ticks=False):
        """
        Author: Jakob Kruse (jakob.kruse@iwr.uni-heidelberg.de)
        Modified by: Maxime Vandegar (maxime.vandegar@slac.stanford.edu)
        """
        
        plt.figure(figsize=(7, 7))
        cmap = plt.get_cmap('tab20c')
        colors = [cmap(4*3), cmap(4*3+1), cmap(4*3+2)]
        
        x0 = torch.zeros(x.shape[0], 2)
        x0[:, 1] = x[:, 0]
        
        x1 = torch.zeros(x.shape[0], 2)
        x1[:, 0] = x0[:, 0] + self.l1 * torch.cos(x[:,1])
        x1[:, 1] = x0[:, 1] + self.l1 * torch.sin(x[:,1])
        
        x2 = torch.zeros(x.shape[0], 2)
        x2[:, 0] = x1[:, 0] + self.l2 * torch.cos(x[:,1] + x[:,2])
        x2[:, 1] = x1[:, 1] + self.l2 * torch.sin(x[:,1] + x[:,2])
        
        x3 = torch.zeros(x.shape[0], 2)
        x3[:, 0] = x2[:, 0] + self.l3 * torch.cos(x[:,1] + x[:,2] + x[:,3])
        x3[:, 1] = x2[:, 1] + self.l3 * torch.sin(x[:,1] + x[:,2] + x[:,3])
        
        plt.axvline(x=0, ls=':', c='gray', linewidth=2)
        
        opts = {'alpha':0.05, 
                'scale':1, 
                'angles':'xy', 
                'scale_units':'xy', 
                'headlength':0, 
                'headaxislength':0, 
                'linewidth':1.0, 
                'rasterized':True}
        
        plt.quiver(x0[:,0], x0[:,1], (x1-x0)[:,0], (x1-x0)[:,1], **{'color': colors[0], **opts})
        plt.quiver(x1[:,0], x1[:,1], (x2-x1)[:,0], (x2-x1)[:,1], **{'color': colors[1], **opts})
        plt.quiver(x2[:,0], x2[:,1], (x3-x2)[:,0], (x3-x2)[:,1], **{'color': colors[2], **opts})
        plt.scatter(x3[:,0], x3[:,1], s=4, c=colors[2], edgecolors=colors[0], linewidths=0.5, alpha=0.5)
        
        '''plt.xlim(-0.2, 2.2); 
        plt.ylim(-1.2, 1.2)'''

        if disable_ticks:
            plt.gca().set_xticks([]); 
            plt.gca().set_yticks([])
            
        plt.tight_layout(pad=0, w_pad=-0.5, h_pad=-0.5)
        
        if savepath is not None:
            plt.savefig(savepath)
        if do_show:
            plt.show()
        else:
            plt.close()


def sample_multivariate_gaussian(size, mean_1, mean_2):
    
    means =  torch.zeros(size, 2)
    means[:, 0] += mean_1
    means[:, 1] += mean_2
    cov =  torch.eye(2)
    
    m = MultivariateNormal(means, cov)
    return m.sample()

# TODO, create class simulator

def sample_p_z(size):
    means =  torch.zeros(size, 2)
    cov =  torch.eye(2)
    
    m = MultivariateNormal(means, cov)
    return m.sample()

def p_y_x(x, y):
    '''
    For a pair (x,y), returns the log probability log(p_{y|x}(y|x)) defined 
    by the simulator.
    '''
    means = (0.3 * x**3 + 0.2 * x**2 + x  + 1)
    cov = torch.eye(2)
    m = MultivariateNormal(means, cov)
    log_prob = m.log_prob(y)
    return log_prob
    
    
def log_p_z(z):
    m = MultivariateNormal(torch.tensor([0.0, 0.0]), torch.eye(2))
    log_prob = m.log_prob(z)
    return log_prob

def log_p_y(z, x):
    means = x * 2  + 1
    cov = torch.eye(2)
    m = MultivariateNormal(means, cov)
    log_prob = m.log_prob(z)
    return log_prob