import os
import numpy
from matplotlib import pyplot as plt
import imageio
import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import Dataset, DataLoader
from torch.distributions.normal import Normal 
from torch.distributions.multivariate_normal import MultivariateNormal
import torch.optim as optim
from torch.autograd import Variable
import numpy as np

       
class RealNVP(nn.Module):

    def __init__(self, z_dim, x_dim, hidden_layer_dim, use_batch_norm = False, use_weight_norm = False, alpha = 1.9, use_clamping=True):
        super().__init__()

        self.d = z_dim//2  # d in the initial paper
        output_dim = z_dim - self.d # D - d in the initial paper
        self.alpha = alpha # alpha for the clamping operator defined in 'GUIDED IMAGE GENERATION WITH
                           #                                             CONDITIONAL INVERTIBLE NEURAL 
                           #                                             NETWORKS'
        self.use_clamping = use_clamping

        if(use_weight_norm):
            # TODO
            pass
        else:
            if(use_batch_norm):
                print('use bacth_norm')
                self.s = nn.Sequential(
                    nn.Linear(self.d + x_dim, hidden_layer_dim),
                    nn.ReLU(),
                    nn.Linear(hidden_layer_dim, hidden_layer_dim),
                    nn.BatchNorm1d(hidden_layer_dim),
                    nn.ReLU(),
                    nn.Linear(hidden_layer_dim, output_dim)
                    )

                self.t = nn.Sequential(
                    nn.Linear(self.d + x_dim, hidden_layer_dim),
                    nn.ReLU(),
                    nn.Linear(hidden_layer_dim, hidden_layer_dim),
                    nn.BatchNorm1d(hidden_layer_dim),
                    nn.ReLU(),
                    nn.Linear(hidden_layer_dim, output_dim)
                    )

            else:
                
                self.s = nn.Sequential(
                    nn.Linear(self.d + x_dim, hidden_layer_dim),
                    nn.ReLU(),
                    nn.Linear(hidden_layer_dim, hidden_layer_dim),
                    nn.ReLU(),
                    nn.Linear(hidden_layer_dim, output_dim)
                    )

                self.t = nn.Sequential(
                    nn.Linear(self.d + x_dim, hidden_layer_dim),
                    nn.ReLU(),
                    nn.Linear(hidden_layer_dim, hidden_layer_dim),
                    nn.ReLU(),
                    nn.Linear(hidden_layer_dim, output_dim)
                    )

    def forward(self, z, x):
        return self.direct_forward(z, x)

    def direct_forward(self, z, x): # z should be renamed to y
        y1_d = z[:,:self.d]        # z should be renamed to y
        
        if x is None: # Non conditional case
            s = self.s(y1_d)       # z should be renamed to y
        else:
            s = self.s(torch.cat((y1_d, x), dim=1))  # z should be renamed to y

        
        s = self.clamp(s)

        if x is None: # Non conditional case
            y_d1_D = torch.mul(z[:,self.d:], torch.exp(s)) + self.t(y1_d)
        else:
            y_d1_D = torch.mul(z[:,self.d:], torch.exp(s)) + self.t(torch.cat((y1_d, x), dim=1))

        return torch.cat(( y1_d, y_d1_D), dim=1), torch.sum(s, dim=1)

    def inverse_forward(self, y, x):   # z should be renamed to y
        z1_d = y[:,:self.d]           # z should be renamed to y
        
        if x is None: # Non conditional case
            z_d1_D = torch.mul(y[:,self.d:] - self.t(y[:,:self.d]), torch.exp( - self.clamp(self.s(y[:,:self.d]))))
        else:
            z_d1_D = torch.mul(y[:,self.d:] - self.t(torch.cat((y[:,:self.d], x),dim=1)), torch.exp( - self.clamp(self.s(torch.cat((y[:,:self.d], x), dim=1)))))

        return torch.cat((z1_d, z_d1_D), dim=1) 

    def clamp(self, s):
        """
        Clamping operator defined in 'GUIDED IMAGE GENERATION WITH CONDITIONAL 
                                      INVERTIBLE NEURAL NETWORKS'
        """
        if(self.use_clamping):
            return (2 * self.alpha / math.pi) * torch.atan(s/self.alpha)
        else:
            return s


class NormalizingFlow(nn.Module):

    def __init__(self, flow_length, z_dim, x_dim, hidden_layer_dim, use_batch_norm = False, use_clamping=True, alpha = 1.9):
        super().__init__()

        self.z_dim = z_dim
        self.layers = nn.Sequential(
            *(RealNVP(z_dim, x_dim,hidden_layer_dim, use_batch_norm = use_batch_norm, use_clamping=use_clamping, alpha=alpha) for _ in range(flow_length)))

    def forward(self, z, x):
        return self.direct_forward(z, x)

    def direct_forward(self, z, x):   # TODO, add batch norm here

        log_jacobians = 0
        y = z


        for layer in self.layers:
            # Every iteration, we swap the first and last variables so that no variable is
            # unchanged
            y = torch.cat((y[:, self.z_dim//2:], y[:, :self.z_dim//2]), dim=1)
            
            y, log = layer(y, x)
            log_jacobians += log
        return y, log_jacobians

    def inverse_forward(self, y, x):  # TODO, add batch norm here

        z = y
        for i in range(len(self.layers)-1, -1, -1):
            z = self.layers[i].inverse_forward(z, x)

            # Every iteration, we swap the first and last variables so that no variable is
            # unchanged
            z = torch.cat((z[:, (self.z_dim - self.z_dim//2):], z[:, :(self.z_dim - self.z_dim//2)]), dim=1)

        return z   
        


