from sklearn import preprocessing
import pandas as pd
import numpy

# Wrapper for different data processors
def Data_Processor(name):
    
    if(name == 'StandardScaler'):
        return StandardScaler()
    elif(name == 'LogScaler'):
        return LogScaler()
    elif(name == 'LogStandardScaler'):
        return LogStandardScaler()
    elif(name == 'CosSinLogStandardScaler'):
        return CosSinLogStandardScaler()
    elif(name == 'MinMaxScaler'):
        return MinMaxScaler()
    else:
        raise ValueError('Unknown Processor Name.')
        
class StandardScaler():
    
    def load_params(self, parameters_file_path):        
        data = pd.read_csv(parameters_file_path)

        self.scaler_x = preprocessing.StandardScaler()
        self.scaler_x.mean_ = data['x_mean'].values.tolist()
        self.scaler_x.scale_ = data['x_scale'].values.tolist()
        
        self.scaler_y = preprocessing.StandardScaler()
        self.scaler_y.mean_ = data['y_mean'].values.tolist()
        self.scaler_y.scale_ = data['y_scale'].values.tolist()
        
    def fit(self, X, Y, save_file_path):
        self.scaler_x = preprocessing.StandardScaler().fit(X)
        self.scaler_y = preprocessing.StandardScaler().fit(Y)
        
        df = pd.DataFrame(data={"x_mean": self.scaler_x.mean_, 
                                "x_scale": self.scaler_x.scale_, 
                                "y_mean": self.scaler_y.mean_,
                                "y_scale": self.scaler_y.scale_})

        df.to_csv(save_file_path, sep=',',index=False)

        
    def pre_process(self, X, Y):
        return self.scaler_x.transform(X), self.scaler_y.transform(Y)
    
    def post_process(self, X, Y):
        processed_X = X * self.scaler_x.scale_ + self.scaler_x.mean_
        processed_Y = Y * self.scaler_y.scale_ + self.scaler_y.mean_

        return processed_X, processed_Y

class LogScaler():
    
    def load_params(self, parameters_file_path):
        raise NotImplementedError("No parameters to load for this Scaler") 
    
    def fit(self, X, Y, save_file_path):
        raise NotImplementedError("No fitting needed for this Scaler")
        
    def _first_pre_processing_step_(self, X, Y):
        X[:, 0] = numpy.log(X[:, 0])
        X[:, 3] = numpy.log(X[:, 3] + 1)

        Y[:, 0] = numpy.log(Y[:, 0])
        Y[:, 3] = numpy.log(Y[:, 3] + 1)
        
        return X, Y
        
    def pre_process(self, X, Y):
        
        X, Y = self._first_pre_processing_step_(X, Y)
        return X, Y
    
    def post_process(self, X, Y):
        
        X[:, 0] = numpy.exp(X[:, 0])
        X[:, 3] = numpy.exp(X[:, 3]) - 1

        Y[:, 0] = numpy.exp(Y[:, 0])
        Y[:, 3] = numpy.exp(Y[:, 3]) - 1

        return X, Y
        
class LogStandardScaler():
    
    def load_params(self, parameters_file_path):
        data = pd.read_csv(parameters_file_path)

        self.scaler_x = preprocessing.StandardScaler()
        self.scaler_x.mean_ = data['x_mean'].values.tolist()
        self.scaler_x.scale_ = data['x_scale'].values.tolist()
        
        self.scaler_y = preprocessing.StandardScaler()
        self.scaler_y.mean_ = data['y_mean'].values.tolist()
        self.scaler_y.scale_ = data['y_scale'].values.tolist() 
    
    def fit(self, X, Y, save_file_path):
        
        X, Y = self._first_pre_processing_step_(X.copy(), Y.copy())
        
        self.scaler_x = preprocessing.StandardScaler().fit(X)
        self.scaler_y = preprocessing.StandardScaler().fit(Y)
        
        df = pd.DataFrame(data={"x_mean": self.scaler_x.mean_, 
                                "x_scale": self.scaler_x.scale_, 
                                "y_mean": self.scaler_y.mean_,
                                "y_scale": self.scaler_y.scale_})

        df.to_csv(save_file_path, sep=',',index=False)
        
    def _first_pre_processing_step_(self, X, Y):
        X[:, 0] = numpy.log(X[:, 0])
        X[:, 3] = numpy.log(X[:, 3] + 1)

        Y[:, 0] = numpy.log(Y[:, 0])
        Y[:, 3] = numpy.log(Y[:, 3] + 1)
        
        return X, Y
        
    def pre_process(self, X, Y):
        
        X, Y = self._first_pre_processing_step_(X, Y)
        
        return self.scaler_x.transform(X), self.scaler_y.transform(Y)
    
    def post_process(self, X, Y):
        
        X = X * self.scaler_x.scale_ + self.scaler_x.mean_
        Y = Y * self.scaler_y.scale_ + self.scaler_y.mean_
        
        X[:, 0] = numpy.exp(X[:, 0])
        X[:, 3] = numpy.exp(X[:, 3]) - 1

        Y[:, 0] = numpy.exp(Y[:, 0])
        Y[:, 3] = numpy.exp(Y[:, 3]) - 1

        return X, Y
    
class CosSinLogStandardScaler():
    
    def load_params(self, parameters_file_path):
        data = pd.read_csv(parameters_file_path)

        self.scaler_x = preprocessing.StandardScaler()
        self.scaler_x.mean_ = data['x_mean'].values.tolist()
        self.scaler_x.scale_ = data['x_scale'].values.tolist()
        
        self.scaler_y = preprocessing.StandardScaler()
        self.scaler_y.mean_ = data['y_mean'].values.tolist()
        self.scaler_y.scale_ = data['y_scale'].values.tolist()
        
    def fit(self, X, Y, save_file_path):
        
        X, Y = self._first_pre_processing_step_(X.copy(), Y.copy())
        
        self.scaler_x = preprocessing.StandardScaler().fit(X)
        self.scaler_y = preprocessing.StandardScaler().fit(Y)
        
        df = pd.DataFrame(data={"x_mean": self.scaler_x.mean_, 
                                "x_scale": self.scaler_x.scale_, 
                                "y_mean": self.scaler_y.mean_,
                                "y_scale": self.scaler_y.scale_})

        df.to_csv(save_file_path, sep=',',index=False)
    
    def _first_pre_processing_step_(self, X, Y):
        
        Processed_X = numpy.zeros((X.shape[0], 5))
        Processed_Y = numpy.zeros((Y.shape[0], 5))
         
        Processed_X[:, 0] = numpy.log(X[:, 0])
        Processed_X[:, 1] = X[:, 1]
        Processed_X[:, 2] = numpy.cos(X[:, 2])
        Processed_X[:, 3] = numpy.sin(X[:, 2])
        Processed_X[:, 4] = numpy.log(X[:, 3] + 1)
        
        Processed_Y[:, 0] = numpy.log(Y[:, 0])
        Processed_Y[:, 1] = Y[:, 1]
        Processed_Y[:, 2] = numpy.cos(Y[:, 2])
        Processed_Y[:, 3] = numpy.sin(Y[:, 2])
        Processed_Y[:, 4] = numpy.log(Y[:, 3] + 1)
        return Processed_X, Processed_Y
        
    def pre_process(self, X, Y):
        
        X, Y = self._first_pre_processing_step_(X, Y)

        return self.scaler_x.transform(X), self.scaler_y.transform(Y)
    
    def post_process(self, X, Y):
        
        X = X * self.scaler_x.scale_ + self.scaler_x.mean_
        Y = Y * self.scaler_y.scale_ + self.scaler_y.mean_
        
        Processed_X = numpy.zeros((X.shape[0], 4))
        Processed_Y = numpy.zeros((Y.shape[0], 4))
        
        Processed_X[:, 0] = numpy.exp(X[:, 0])
        Processed_X[:, 1] = X[:, 1]
        Processed_X[:, 2] = numpy.arctan2(X[:, 3], X[:, 2])
        Processed_X[:, 2][Processed_X[:, 2] < 0] = Processed_X[:, 2][Processed_X[:, 2] < 0] + 2 * numpy.pi
        Processed_X[:, 3] = numpy.exp(X[:, 4]) - 1

        Processed_Y[:, 0] = numpy.exp(Y[:, 0])
        Processed_Y[:, 1] = Y[:, 1]
        Processed_Y[:, 2] = numpy.arctan2(Y[:, 3], Y[:, 2])
        Processed_Y[:, 2][Processed_Y[:, 2] < 0] = Processed_Y[:, 2][Processed_Y[:, 2] < 0] + 2 * numpy.pi
        Processed_Y[:, 3] = numpy.exp(Y[:, 4]) - 1

        return Processed_X, Processed_Y
        
    
class MinMaxScaler():
    
    def __init__(self, data, X, Y):
        raise NotImplementedError("Not implemented yet")
