import os
import numpy
from matplotlib import pyplot as plt
import imageio

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import Dataset, DataLoader
from torch.distributions.normal import Normal 
from torch.distributions.multivariate_normal import MultivariateNormal
import torch.optim as optim
from torch.autograd import Variable
import numpy as np


# Latent distribution
def sample_p_z(size, dim=2):
    means =  torch.zeros(size, dim)
    cov =  torch.eye(dim)
    
    m = MultivariateNormal(means, cov)
    return m.sample()

def log_p_z(z):
    dim = z.shape[1]
    m = MultivariateNormal(torch.zeros(dim), torch.eye(dim))
    log_prob = m.log_prob(z)
    return log_prob


def plot_density_given_x( density_function, model, log_p_z, x=0, x_dim=0, X_LIMS = (0, 20), Y_LIMS = (0, 20), Nb_point_per_dimension = 100,directory= None, figName=None):
    '''
    Given a conditional density function and an approximation of this density, plot
    or save the densities given x
    
    Arguments:
        ----------
        - 'density function': The true conditional density
        - 'model': An approximation of the density function.
        - 'x': The value on which conditioning should be done.
        - 'x_dim': The dimension of the value x.
        - 'X_LIMS':
        - 'Y_LIMS':
        - 'Nb_point_per_dimension':
        - 'directory': The directory name to save the plot. If set to None, the plot is shown but not saved.
        - 'figName': The file name to save the plot. If set to None, the plot is shown but not saved.
    '''
    
    # Format the limits so that python indexing matches with plt indexing - pixel (0,0) in the bottom left
    X_LIMS = (X_LIMS[0], X_LIMS[1])
    Y_LIMS = (Y_LIMS[1], Y_LIMS[0])

    x1 = np.linspace(*X_LIMS, Nb_point_per_dimension)
    x2 = np.linspace(*Y_LIMS, Nb_point_per_dimension)
    x1, x2 = np.meshgrid(x1, x2)
    x1 = x1.ravel()
    x2 = x2.ravel()

    z = np.c_[x1, x2] # z should be renamed to y
    z = torch.FloatTensor(z) # z should be renamed to y
    
    z_, log_jacobian = model(z, torch.zeros((z.shape[0], x_dim)) + x) # z should be renamed to y
    estimated_density = torch.exp(log_p_z(z_) + log_jacobian)
    estimated_density = estimated_density.data.numpy().reshape(Nb_point_per_dimension, Nb_point_per_dimension)

    fig = plt.figure(figsize=(14, 14))
    
    if(density_function):
        true_density = torch.exp(density_function(z, torch.zeros((z.shape[0], x_dim)) + x).reshape(Nb_point_per_dimension,Nb_point_per_dimension))
        
        
        ax = fig.add_subplot(1,2,1)
        ax.imshow(true_density, extent=([X_LIMS[0], X_LIMS[1], Y_LIMS[1], Y_LIMS[0]]), cmap="summer")
        ax.set_title("True density")
        plt.ylabel('y_2')
        plt.xlabel('y_1')
    
    
    ax = fig.add_subplot(1,2,2)
    ax.imshow(estimated_density, extent=([X_LIMS[0], X_LIMS[1], Y_LIMS[1], Y_LIMS[0]]), cmap="summer")
    ax.set_title("Estimated density")
    plt.ylabel('y_2')
    plt.xlabel('y_1')
        
    if(figName and directory):
        if not os.path.exists(directory):
            os.mkdir(directory)
        fig.savefig(os.path.join(directory + figName))
    else:
        plt.show()
    plt.close()


def makeGIF():
    #
    # TODO, change the sorting function, not good
    #

    images = []
    file_names = sorted((fn for fn in os.listdir('./Results/density_given_x_0/') if fn.endswith('.png')))
    for filename in file_names:
        images.append(imageio.imread('Results/density_given_x_0/' + filename))

    imageio.mimsave('./Results/density_given_x_0/movie.gif', images)

    images = []
    file_names = sorted((fn for fn in os.listdir('./Results/density_given_x_3/') if fn.endswith('.png')))
    for filename in file_names:
        images.append(imageio.imread('Results/density_given_x_3/' + filename))

    imageio.mimsave('./Results/density_given_x_3/movie.gif', images)


def calc_summary_stats(y):

    #
    # Todo, comment and generalize
    #
    
    y = y.data.numpy()
    
    my_1 = y.mean(axis=0)[0]
    my_2 = y.mean(axis=0)[1]
    
    log_var_1 = numpy.log(numpy.var(y[:, 0])+1)
    log_var_2 = numpy.log(numpy.var(y[:, 1])+1)
    
    # cross correlation coefficient
    ccxy = numpy.dot(y[:, 0], y[:, 1]) / (y.shape[0]-1)
    
    return numpy.array([my_1, my_2, log_var_1, log_var_2, ccxy])


def diag_plot(samples, true_ps, savepath=None, lims=np.asarray([-5, 5])):

    '''
    Code inspired from https://github.com/gpapamak/snl
    '''

    plt.figure(figsize=(12,12))

    n_dim = samples.shape[1]
    n_bins = int(np.sqrt(samples.shape[0]))

    lims = np.tile(lims, [n_dim, 1]) if lims.ndim == 1 else lims

    for i in range(n_dim):
        for j in range(i, n_dim):

            if i == j:

                ax = plt.subplot(n_dim, n_dim, i*n_dim+j+1)
                ax.hist(samples[:, i], bins=n_bins, density=True, rasterized=True, range=lims[i], color='k')
                '''ax.set_xlim(lims[i])'''
                ax.set_ylim([0.0, ax.get_ylim()[1]])
                ax.tick_params(axis='x', which='both', top=False, bottom=False, labelbottom=False)
                ax.tick_params(axis='y', which='both', left=False, right=False, labelleft=False)
                ax.vlines(true_ps[i], 0, ax.get_ylim()[1], color='r')

            else:


                ax = plt.subplot(n_dim, n_dim, i*n_dim+j+1)
                ax.scatter(samples[:, j], samples[:, i], c='k', s=3, marker='o', cmap='binary', edgecolors='none', rasterized=True)
                ax.tick_params(axis='x', which='both', top=False, bottom=False, labelbottom=False)
                ax.tick_params(axis='y', which='both', left=False, right=False, labelleft=False)
                ax.set_xlim(lims[j])
                ax.set_ylim(lims[i])
                ax.scatter(true_ps[j], true_ps[i], c='r', s=50, marker='o', edgecolors='none')

    if savepath is not None:
        plt.savefig(savepath)
    plt.show()

def plot_surrogate_data(true_data, learned_data, lims=np.asarray([-5, 5]), savepath=None, fontsize=18, fontweight='bold', color='cornflowerblue'):
    
    plt.figure(figsize=(15,15))

    n_dim = true_data.shape[1]
    n_bins = int(np.sqrt(true_data.shape[0]))

    lims = np.tile(lims, [n_dim, 1]) if lims.ndim == 1 else lims
    
    for i in range(n_dim):
        for j in range(i, n_dim):

            if i == j:

                ax = plt.subplot(n_dim+1, n_dim + 1, i*(n_dim+1)+j+1 +0)
                ax.hist(true_data[:, i], bins=n_bins, density=True, rasterized=True, range=lims[i], color='skyblue')
                #ax.hist(true_data[:, i], bins=n_bins, density=True, rasterized=True, range=lims[i], color='b', histtype='step')
                ax.hist(learned_data[:, i], bins=n_bins, density=True, rasterized=True, range=lims[i], color='k', histtype='step')

                '''ax.set_xlim(lims[i])'''
                ax.set_ylim([0.0, ax.get_ylim()[1]])
                ax.tick_params(axis='x', which='both', top=False, bottom=False, labelbottom=False)
                ax.tick_params(axis='y', which='both', left=False, right=False, labelleft=False)
                ax.set_xlabel(f'$y_{i+1}$', fontsize=fontsize, fontweight=fontweight)
            
            else:
                
                


                ax = plt.subplot(n_dim+1, n_dim + 1, i*(n_dim+1)+j+1 +1)
                ax.scatter(learned_data[:, i], learned_data[:, j], c=color, s=3, marker='o', cmap='binary', edgecolors='none', rasterized=True, alpha=0.05)
                #  ax.hist2d(y[:, j], y[:, i], bins=n_bins, range = y_lims )#, c='k', s=3, marker='o', cmap='binary', edgecolors='none', rasterized=True)

                ax.tick_params(axis='x', which='both', top=False, bottom=False, labelbottom=False)
                ax.tick_params(axis='y', which='both', left=False, right=False, labelleft=False)
                ax.set_xlim(lims[j])
                ax.set_ylim(lims[i])
                if j == (i+1):
                    ax.set_ylabel(f'$y_{j}$', fontsize=fontsize, fontweight=fontweight)
                    ax.set_xlabel(f'$y_{i+2}$', fontsize=fontsize, fontweight=fontweight)

        for j in range(0, i):
            ax = plt.subplot(n_dim+1, n_dim + 1, (i+1)*(n_dim+1)+j+1 +0)
            ax.scatter(learned_data[:, i], learned_data[:, j], c=color, s=3, marker='o', cmap='binary', edgecolors='none', rasterized=True, alpha=0.05)
            #  ax.hist2d(y[:, j], y[:, i], bins=n_bins, range = y_lims )#, c='k', s=3, marker='o', cmap='binary', edgecolors='none', rasterized=True)

            ax.tick_params(axis='x', which='both', top=False, bottom=False, labelbottom=False)
            ax.tick_params(axis='y', which='both', left=False, right=False, labelleft=False)
            ax.set_xlim(lims[j])
            ax.set_ylim(lims[i])
            
            if i == (n_dim - 1):
                ax.set_xlabel(f'$y_{j+1}$', fontsize=fontsize, fontweight=fontweight)
            if j == (i - 1):
                ax.set_ylabel(f'$y_{i+1}$', fontsize=fontsize, fontweight=fontweight)
                ax.yaxis.set_label_position("right")

    if savepath is not None:
        plt.savefig(savepath)
    plt.show()
    